@@ -16,171 +16,30 @@ from typing import (
     Dict,
     Iterable,
     List,
+    MutableMapping,
     NamedTuple,
     Optional,
     Set,
-    Tuple,
     cast,
 )
+from weakref import WeakKeyDictionary
 
-from dbus_next import BusType, Message, MessageType, Variant
-from dbus_next.aio.message_bus import MessageBus
-from typing_extensions import Literal, TypedDict
+from dbus_fast import BusType, Message, MessageType, Variant, unpack_variants
+from dbus_fast.aio.message_bus import MessageBus
 
-from ...exc import BleakError
+from ...exc import BleakDBusError, BleakError
 from ..service import BleakGATTServiceCollection
 from . import defs
 from .advertisement_monitor import AdvertisementMonitor, OrPatternLike
 from .characteristic import BleakGATTCharacteristicBlueZDBus
+from .defs import Device1, GattService1, GattCharacteristic1, GattDescriptor1
 from .descriptor import BleakGATTDescriptorBlueZDBus
 from .service import BleakGATTServiceBlueZDBus
 from .signals import MatchRules, add_match
-from .utils import assert_reply, unpack_variants
+from .utils import assert_reply, get_dbus_authenticator
 
 logger = logging.getLogger(__name__)
 
-
-# D-Bus properties for interfaces
-# https://github.com/bluez/bluez/blob/master/doc/adapter-api.txt
-
-
-class Adapter1(TypedDict):
-    Address: str
-    Name: str
-    Alias: str
-    Class: int
-    Powered: bool
-    Discoverable: bool
-    Pairable: bool
-    PairableTimeout: int
-    DiscoverableTimeout: int
-    Discovering: int
-    UUIDs: List[str]
-    Modalias: str
-    Roles: List[str]
-    ExperimentalFeatures: List[str]
-
-
-# https://github.com/bluez/bluez/blob/master/doc/advertisement-monitor-api.txt
-
-
-class AdvertisementMonitor1(TypedDict):
-    Type: str
-    RSSILowThreshold: int
-    RSSIHighThreshold: int
-    RSSILowTimeout: int
-    RSSIHighTimeout: int
-    RSSISamplingPeriod: int
-    Patterns: List[Tuple[int, int, bytes]]
-
-
-class AdvertisementMonitorManager1(TypedDict):
-    SupportedMonitorTypes: List[str]
-    SupportedFeatures: List[str]
-
-
-# https://github.com/bluez/bluez/blob/master/doc/battery-api.txt
-
-
-class Battery1(TypedDict):
-    SupportedMonitorTypes: List[str]
-    SupportedFeatures: List[str]
-
-
-# https://github.com/bluez/bluez/blob/master/doc/device-api.txt
-
-
-class Device1(TypedDict):
-    Address: str
-    AddressType: str
-    Name: str
-    Icon: str
-    Class: int
-    Appearance: int
-    UUIDs: List[str]
-    Paired: bool
-    Bonded: bool
-    Connected: bool
-    Trusted: bool
-    Blocked: bool
-    WakeAllowed: bool
-    Alias: str
-    Adapter: str
-    LegacyPairing: bool
-    Modalias: str
-    RSSI: int
-    TxPower: int
-    ManufacturerData: Dict[int, bytes]
-    ServiceData: Dict[str, bytes]
-    ServicesResolved: bool
-    AdvertisingFlags: bytes
-    AdvertisingData: Dict[int, bytes]
-
-
-# https://github.com/bluez/bluez/blob/master/doc/gatt-api.txt
-
-
-class GattService1(TypedDict):
-    UUID: str
-    Primary: bool
-    Device: str
-    Includes: List[str]
-    # Handle is server-only and not available in Bleak
-
-
-class GattCharacteristic1(TypedDict):
-    UUID: str
-    Service: str
-    Value: bytes
-    WriteAcquired: bool
-    NotifyAcquired: bool
-    Notifying: bool
-    Flags: List[
-        Literal[
-            "broadcast",
-            "read",
-            "write-without-response",
-            "write",
-            "notify",
-            "indicate",
-            "authenticated-signed-writes",
-            "extended-properties",
-            "reliable-write",
-            "writable-auxiliaries",
-            "encrypt-read",
-            "encrypt-write",
-            # "encrypt-notify" and "encrypt-indicate" are server-only
-            "encrypt-authenticated-read",
-            "encrypt-authenticated-write",
-            # "encrypt-authenticated-notify", "encrypt-authenticated-indicate",
-            # "secure-read", "secure-write", "secure-notify", "secure-indicate"
-            # are server-only
-            "authorize",
-        ]
-    ]
-    MTU: int
-    # Handle is server-only and not available in Bleak
-
-
-class GattDescriptor1(TypedDict):
-    UUID: str
-    Characteristic: str
-    Value: bytes
-    Flags: List[
-        Literal[
-            "read",
-            "write",
-            "encrypt-read",
-            "encrypt-write",
-            "encrypt-authenticated-read",
-            "encrypt-authenticated-write",
-            # "secure-read" and "secure-write" are server-only and not available in Bleak
-            "authorize",
-        ]
-    ]
-    # Handle is server-only and not available in Bleak
-
-
 AdvertisementCallback = Callable[[str, Device1], None]
 """
 A callback that is called when advertisement data is received.
@@ -206,9 +65,29 @@ class CallbackAndState(NamedTuple):
     The D-Bus object path of the adapter associated with the callback.
     """
 
-    seen_devices: Set[str]
+
+DeviceRemovedCallback = Callable[[str], None]
+"""
+A callback that is called when a device is removed from BlueZ.
+
+Args:
+    arg0: The D-Bus object path of the device.
+"""
+
+
+class DeviceRemovedCallbackAndState(NamedTuple):
     """
-    Set of device D-Bus object paths that have been "seen" already by this callback.
+    Encapsulates an :data:`DeviceRemovedCallback` and some state.
+    """
+
+    callback: DeviceRemovedCallback
+    """
+    The callback.
+    """
+
+    adapter_path: str
+    """
+    The D-Bus object path of the adapter associated with the callback.
     """
 
 
@@ -273,9 +152,24 @@ class BlueZManager:
         # dict of object path: dict of interface name: dict of property name: property value
         self._properties: Dict[str, Dict[str, Dict[str, Any]]] = {}
 
+        # set of available adapters for quick lookup
+        self._adapters: Set[str] = set()
+
+        # The BlueZ APIs only maps children to parents, so we need to keep maps
+        # to quickly find the children of a parent D-Bus object.
+
+        # map of device d-bus object paths to set of service d-bus object paths
+        self._service_map: Dict[str, Set[str]] = {}
+        # map of service d-bus object paths to set of characteristic d-bus object paths
+        self._characteristic_map: Dict[str, Set[str]] = {}
+        # map of characteristic d-bus object paths to set of descriptor d-bus object paths
+        self._descriptor_map: Dict[str, Set[str]] = {}
+
         self._advertisement_callbacks: List[CallbackAndState] = []
+        self._device_removed_callbacks: List[DeviceRemovedCallbackAndState] = []
         self._device_watchers: Set[DeviceWatcher] = set()
         self._condition_callbacks: Set[Callable] = set()
+        self._services_cache: Dict[str, BleakGATTServiceCollection] = {}
 
     async def async_init(self):
         """
@@ -288,10 +182,12 @@ class BlueZManager:
             if self._bus and self._bus.connected:
                 return
 
+            self._services_cache = {}
+
             # We need to create a new MessageBus each time as
             # dbus-next will destory the underlying file descriptors
             # when the previous one is closed in its finalizer.
-            bus = MessageBus(bus_type=BusType.SYSTEM)
+            bus = MessageBus(bus_type=BusType.SYSTEM, auth=get_dbus_authenticator())
             await bus.connect()
 
             try:
@@ -336,11 +232,47 @@ class BlueZManager:
                 )
                 assert_reply(reply)
 
-                # dictionary is replaced in case AddInterfaces was received first
-                self._properties = {
-                    path: unpack_variants(interfaces)
-                    for path, interfaces in reply.body[0].items()
-                }
+                # dictionaries are cleared in case AddInterfaces was received first
+                # or there was a bus reset and we are reconnecting
+                self._properties.clear()
+                self._service_map.clear()
+                self._characteristic_map.clear()
+                self._descriptor_map.clear()
+
+                for path, interfaces in reply.body[0].items():
+                    props = unpack_variants(interfaces)
+                    self._properties[path] = props
+
+                    if defs.ADAPTER_INTERFACE in props:
+                        self._adapters.add(path)
+
+                    service_props = cast(
+                        GattService1, props.get(defs.GATT_SERVICE_INTERFACE)
+                    )
+
+                    if service_props:
+                        self._service_map.setdefault(
+                            service_props["Device"], set()
+                        ).add(path)
+
+                    char_props = cast(
+                        GattCharacteristic1,
+                        props.get(defs.GATT_CHARACTERISTIC_INTERFACE),
+                    )
+
+                    if char_props:
+                        self._characteristic_map.setdefault(
+                            char_props["Service"], set()
+                        ).add(path)
+
+                    desc_props = cast(
+                        GattDescriptor1, props.get(defs.GATT_DESCRIPTOR_INTERFACE)
+                    )
+
+                    if desc_props:
+                        self._descriptor_map.setdefault(
+                            desc_props["Characteristic"], set()
+                        ).add(path)
 
                 logger.debug(f"initial properties: {self._properties}")
 
@@ -352,11 +284,34 @@ class BlueZManager:
             # Everything is setup, so save the bus
             self._bus = bus
 
+    def get_default_adapter(self) -> str:
+        """
+        Gets the D-Bus object path of of the first powered Bluetooth adapter.
+
+        Returns:
+            Name of the first found powered adapter on the system, i.e. "/org/bluez/hciX".
+
+        Raises:
+            BleakError:
+                if there are no Bluetooth adapters or if none of the adapters are powered
+        """
+        if not any(self._adapters):
+            raise BleakError("No Bluetooth adapters found.")
+
+        for adapter_path in self._adapters:
+            if cast(
+                defs.Adapter1, self._properties[adapter_path][defs.ADAPTER_INTERFACE]
+            )["Powered"]:
+                return adapter_path
+
+        raise BleakError("No powered Bluetooth adapters found.")
+
     async def active_scan(
         self,
         adapter_path: str,
         filters: Dict[str, Variant],
-        callback: AdvertisementCallback,
+        advertisement_callback: AdvertisementCallback,
+        device_removed_callback: DeviceRemovedCallback,
     ) -> Callable[[], Coroutine]:
         """
         Configures the advertisement data filters and starts scanning.
@@ -364,7 +319,10 @@ class BlueZManager:
         Args:
             adapter_path: The D-Bus object path of the adapter to use for scanning.
             filters: A dictionary of filters to pass to ``SetDiscoveryFilter``.
-            callback: A callable that will be called when new advertisement data is received.
+            advertisement_callback:
+                A callable that will be called when new advertisement data is received.
+            device_removed_callback:
+                A callable that will be called when a device is removed from BlueZ.
 
         Returns:
             An async function that is used to stop scanning and remove the filters.
@@ -376,9 +334,14 @@ class BlueZManager:
             if adapter_path not in self._properties:
                 raise BleakError(f"adapter '{adapter_path.split('/')[-1]}' not found")
 
-            callback_and_state = CallbackAndState(callback, adapter_path, set())
+            callback_and_state = CallbackAndState(advertisement_callback, adapter_path)
             self._advertisement_callbacks.append(callback_and_state)
 
+            device_removed_callback_and_state = DeviceRemovedCallbackAndState(
+                device_removed_callback, adapter_path
+            )
+            self._device_removed_callbacks.append(device_removed_callback_and_state)
+
             try:
                 # Apply the filters
                 reply = await self._bus.call(
@@ -405,6 +368,14 @@ class BlueZManager:
                 assert_reply(reply)
 
                 async def stop() -> None:
+                    # need to remove callbacks first, otherwise we get TxPower
+                    # and RSSI properties removed during stop which causes
+                    # incorrect advertisement data callbacks
+                    self._advertisement_callbacks.remove(callback_and_state)
+                    self._device_removed_callbacks.remove(
+                        device_removed_callback_and_state
+                    )
+
                     async with self._bus_lock:
                         reply = await self._bus.call(
                             Message(
@@ -414,34 +385,39 @@ class BlueZManager:
                                 member="StopDiscovery",
                             )
                         )
-                        assert_reply(reply)
 
-                        # remove the filters
-                        reply = await self._bus.call(
-                            Message(
-                                destination=defs.BLUEZ_SERVICE,
-                                path=adapter_path,
-                                interface=defs.ADAPTER_INTERFACE,
-                                member="SetDiscoveryFilter",
-                                signature="a{sv}",
-                                body=[{}],
+                        try:
+                            assert_reply(reply)
+                        except BleakDBusError as ex:
+                            if ex.dbus_error != "org.bluez.Error.NotReady":
+                                raise
+                        else:
+                            # remove the filters
+                            reply = await self._bus.call(
+                                Message(
+                                    destination=defs.BLUEZ_SERVICE,
+                                    path=adapter_path,
+                                    interface=defs.ADAPTER_INTERFACE,
+                                    member="SetDiscoveryFilter",
+                                    signature="a{sv}",
+                                    body=[{}],
+                                )
                             )
-                        )
-                        assert_reply(reply)
-
-                        self._advertisement_callbacks.remove(callback_and_state)
+                            assert_reply(reply)
 
                 return stop
             except BaseException:
-                # if starting scanning failed, don't leak the callback
+                # if starting scanning failed, don't leak the callbacks
                 self._advertisement_callbacks.remove(callback_and_state)
+                self._device_removed_callbacks.remove(device_removed_callback_and_state)
                 raise
 
     async def passive_scan(
         self,
         adapter_path: str,
         filters: List[OrPatternLike],
-        callback: AdvertisementCallback,
+        advertisement_callback: AdvertisementCallback,
+        device_removed_callback: DeviceRemovedCallback,
     ) -> Callable[[], Coroutine]:
         """
         Configures the advertisement data filters and starts scanning.
@@ -449,7 +425,10 @@ class BlueZManager:
         Args:
             adapter_path: The D-Bus object path of the adapter to use for scanning.
             filters: A list of "or patterns" to pass to ``org.bluez.AdvertisementMonitor1``.
-            callback: A callable that will be called when new advertisement data is received.
+            advertisement_callback:
+                A callable that will be called when new advertisement data is received.
+            device_removed_callback:
+                A callable that will be called when a device is removed from BlueZ.
 
         Returns:
             An async function that is used to stop scanning and remove the filters.
@@ -461,9 +440,14 @@ class BlueZManager:
             if adapter_path not in self._properties:
                 raise BleakError(f"adapter '{adapter_path.split('/')[-1]}' not found")
 
-            callback_and_state = CallbackAndState(callback, adapter_path, set())
+            callback_and_state = CallbackAndState(advertisement_callback, adapter_path)
             self._advertisement_callbacks.append(callback_and_state)
 
+            device_removed_callback_and_state = DeviceRemovedCallbackAndState(
+                device_removed_callback, adapter_path
+            )
+            self._device_removed_callbacks.append(device_removed_callback_and_state)
+
             try:
                 monitor = AdvertisementMonitor(filters)
 
@@ -497,6 +481,14 @@ class BlueZManager:
                 self._bus.export(monitor_path, monitor)
 
                 async def stop():
+                    # need to remove callbacks first, otherwise we get TxPower
+                    # and RSSI properties removed during stop which causes
+                    # incorrect advertisement data callbacks
+                    self._advertisement_callbacks.remove(callback_and_state)
+                    self._device_removed_callbacks.remove(
+                        device_removed_callback_and_state
+                    )
+
                     async with self._bus_lock:
                         self._bus.unexport(monitor_path, monitor)
 
@@ -512,13 +504,12 @@ class BlueZManager:
                         )
                         assert_reply(reply)
 
-                        self._advertisement_callbacks.remove(callback_and_state)
-
                 return stop
 
             except BaseException:
-                # if starting scanning failed, don't leak the callback
+                # if starting scanning failed, don't leak the callbacks
                 self._advertisement_callbacks.remove(callback_and_state)
+                self._device_removed_callbacks.remove(device_removed_callback_and_state)
                 raise
 
     def add_device_watcher(
@@ -561,58 +552,78 @@ class BlueZManager:
         """
         self._device_watchers.remove(watcher)
 
-    async def get_services(self, device_path: str) -> BleakGATTServiceCollection:
+    async def get_services(
+        self, device_path: str, use_cached: bool, requested_services: Optional[Set[str]]
+    ) -> BleakGATTServiceCollection:
         """
         Builds a new :class:`BleakGATTServiceCollection` from the current state.
 
         Args:
-            device_path: The D-Bus object path of the Bluetooth device.
+            device_path:
+                The D-Bus object path of the Bluetooth device.
+            use_cached:
+                When ``True`` if there is a cached :class:`BleakGATTServiceCollection`,
+                the method will not wait for ``"ServicesResolved"`` to become true
+                and instead return the cached service collection immediately.
+            requested_services:
+                When given, only return services with UUID that is in the list
+                of requested services.
 
         Returns:
             A new :class:`BleakGATTServiceCollection`.
         """
-        await self._wait_condition(device_path, "ServicesResolved", True)
+        if use_cached:
+            services = self._services_cache.get(device_path)
+            if services is not None:
+                logger.debug("Using cached services for %s", device_path)
+                return services
+
+        await self._wait_for_services_discovery(device_path)
 
         services = BleakGATTServiceCollection()
 
-        for service_path, service_ifaces in self._properties.items():
+        for service_path in self._service_map.get(device_path, set()):
+            service_props = cast(
+                GattService1,
+                self._properties[service_path][defs.GATT_SERVICE_INTERFACE],
+            )
+
+            service = BleakGATTServiceBlueZDBus(service_props, service_path)
+
             if (
-                not service_path.startswith(device_path)
-                or defs.GATT_SERVICE_INTERFACE not in service_ifaces
+                requested_services is not None
+                and service.uuid not in requested_services
             ):
                 continue
 
-            service = BleakGATTServiceBlueZDBus(
-                service_ifaces[defs.GATT_SERVICE_INTERFACE], service_path
-            )
-
             services.add_service(service)
 
-            for char_path, char_ifaces in self._properties.items():
-                if (
-                    not char_path.startswith(service_path)
-                    or defs.GATT_CHARACTERISTIC_INTERFACE not in char_ifaces
-                ):
-                    continue
+            for char_path in self._characteristic_map.get(service_path, set()):
+                char_props = cast(
+                    GattCharacteristic1,
+                    self._properties[char_path][defs.GATT_CHARACTERISTIC_INTERFACE],
+                )
 
                 char = BleakGATTCharacteristicBlueZDBus(
-                    char_ifaces[defs.GATT_CHARACTERISTIC_INTERFACE],
+                    char_props,
                     char_path,
                     service.uuid,
                     service.handle,
+                    # "MTU" property was added in BlueZ 5.62, otherwise fall
+                    # back to minimum MTU according to Bluetooth spec.
+                    char_props.get("MTU", 23) - 3,
                 )
 
                 services.add_characteristic(char)
 
-                for desc_path, desc_ifaces in self._properties.items():
-                    if (
-                        not desc_path.startswith(char_path)
-                        or defs.GATT_DESCRIPTOR_INTERFACE not in desc_ifaces
-                    ):
-                        continue
+                for desc_path in self._descriptor_map.get(char_path, set()):
+                    desc_props = cast(
+                        GattDescriptor1,
+                        self._properties[desc_path][defs.GATT_DESCRIPTOR_INTERFACE],
+                    )
 
                     desc = BleakGATTDescriptorBlueZDBus(
-                        desc_ifaces[defs.GATT_DESCRIPTOR_INTERFACE],
+                        desc_props,
                         desc_path,
                         char.uuid,
                         char.handle,
@@ -620,6 +631,8 @@ class BlueZManager:
 
                     services.add_descriptor(desc)
 
+        self._services_cache[device_path] = services
+
         return services
 
     def get_device_name(self, device_path: str) -> str:
@@ -634,6 +647,44 @@ class BlueZManager:
         """
         return self._properties[device_path][defs.DEVICE_INTERFACE]["Name"]
 
+    def is_connected(self, device_path: str) -> bool:
+        """
+        Gets the value of the "Connected" property for a device.
+
+        Args:
+            device_path: The D-Bus object path of the device.
+
+        Returns:
+            The current property value.
+        """
+        try:
+            return self._properties[device_path][defs.DEVICE_INTERFACE]["Connected"]
+        except KeyError:
+            return False
+
+    async def _wait_for_services_discovery(self, device_path: str) -> None:
+        """
+        Waits for the device services to be discovered.
+
+        If a disconnect happens before the completion a BleakError exception is raised.
+        """
+        services_discovered_wait_task = asyncio.create_task(
+            self._wait_condition(device_path, "ServicesResolved", True)
+        )
+        device_disconnected_wait_task = asyncio.create_task(
+            self._wait_condition(device_path, "Connected", False)
+        )
+        done, pending = await asyncio.wait(
+            {services_discovered_wait_task, device_disconnected_wait_task},
+            return_when=asyncio.FIRST_COMPLETED,
+        )
+
+        for p in pending:
+            p.cancel()
+
+        if device_disconnected_wait_task in done:
+            raise BleakError("failed to discover services, device disconnected")
+
     async def _wait_condition(
         self, device_path: str, property_name: str, property_value: Any
     ) -> None:
@@ -670,14 +721,18 @@ class BlueZManager:
 
     def _parse_msg(self, message: Message):
         """
-        Handles callbacks from dbus_next.
+        Handles callbacks from dbus_fast.
         """
 
         if message.message_type != MessageType.SIGNAL:
             return
 
         logger.debug(
-            f"received D-Bus signal: {message.interface}.{message.member} ({message.path}): {message.body}"
+            "received D-Bus signal: %s.%s (%s): %s",
+            message.interface,
+            message.member,
+            message.path,
+            message.body,
         )
 
         # type hints
@@ -695,12 +750,31 @@ class BlueZManager:
                 unpacked_props = unpack_variants(props)
                 self._properties.setdefault(obj_path, {})[interface] = unpacked_props
 
+                if interface == defs.GATT_SERVICE_INTERFACE:
+                    service_props = cast(GattService1, unpacked_props)
+                    self._service_map.setdefault(service_props["Device"], set()).add(
+                        obj_path
+                    )
+                elif interface == defs.GATT_CHARACTERISTIC_INTERFACE:
+                    char_props = cast(GattCharacteristic1, unpacked_props)
+                    self._characteristic_map.setdefault(
+                        char_props["Service"], set()
+                    ).add(obj_path)
+                elif interface == defs.GATT_DESCRIPTOR_INTERFACE:
+                    desc_props = cast(GattDescriptor1, unpacked_props)
+                    self._descriptor_map.setdefault(
+                        desc_props["Characteristic"], set()
+                    ).add(obj_path)
+
+                elif interface == defs.ADAPTER_INTERFACE:
+                    self._adapters.add(obj_path)
+
                 # If this is a device and it has advertising data properties,
                 # then it should mean that this device just started advertising.
                 # Previously, we just relied on RSSI updates to determine if
                 # a device was actually advertising, but we were missing "slow"
                 # devices that only advertise once and then go to sleep for a while.
-                if interface == defs.DEVICE_INTERFACE:
+                elif interface == defs.DEVICE_INTERFACE:
                     self._run_advertisement_callbacks(
                         obj_path, cast(Device1, unpacked_props), unpacked_props.keys()
                     )
@@ -708,7 +782,36 @@ class BlueZManager:
             obj_path, interfaces = message.body
 
             for interface in interfaces:
-                del self._properties[obj_path][interface]
+                try:
+                    del self._properties[obj_path][interface]
+                except KeyError:
+                    pass
+
+                if interface == defs.ADAPTER_INTERFACE:
+                    try:
+                        self._adapters.remove(obj_path)
+                    except KeyError:
+                        pass
+                elif interface == defs.DEVICE_INTERFACE:
+                    self._services_cache.pop(obj_path, None)
+                    try:
+                        del self._service_map[obj_path]
+                    except KeyError:
+                        pass
+
+                    for callback, adapter_path in self._device_removed_callbacks:
+                        if obj_path.startswith(adapter_path):
+                            callback(obj_path)
+                elif interface == defs.GATT_SERVICE_INTERFACE:
+                    try:
+                        del self._characteristic_map[obj_path]
+                    except KeyError:
+                        pass
+                elif interface == defs.GATT_CHARACTERISTIC_INTERFACE:
+                    try:
+                        del self._descriptor_map[obj_path]
+                    except KeyError:
+                        pass
         elif message.member == "PropertiesChanged":
             assert message.path is not None
 
@@ -717,7 +820,7 @@ class BlueZManager:
             try:
                 self_interface = self._properties[message.path][interface]
             except KeyError:
-                # This can happen during initialization. The "PropertyChanged"
+                # This can happen during initialization. The "PropertiesChanged"
                 # handler is attached before "GetManagedObjects" is called
                 # and so self._properties may not yet be populated.
                 # This is not a problem. We just discard the property value
@@ -729,7 +832,12 @@ class BlueZManager:
                 self_interface.update(unpack_variants(changed))
 
                 for name in invalidated:
-                    del self_interface[name]
+                    try:
+                        del self_interface[name]
+                    except KeyError:
+                        # sometimes there BlueZ tries to remove properties
+                        # that were never added
+                        pass
 
                 # then call any callbacks so they will be called with the
                 # updated state
@@ -776,38 +884,28 @@ class BlueZManager:
             device: The current D-Bus properties of the device.
             changed: A list of properties that have changed since the last call.
         """
-        for (
-            callback,
-            adapter_path,
-            seen_devices,
-        ) in self._advertisement_callbacks:
+        for (callback, adapter_path) in self._advertisement_callbacks:
             # filter messages from other adapters
-            if not device_path.startswith(adapter_path):
+            if adapter_path != device["Adapter"]:
                 continue
 
-            first_time_seen = False
+            callback(device_path, device.copy())
 
-            if device_path not in seen_devices:
-                first_time_seen = True
-                seen_devices.add(device_path)
 
-            # Only do advertising data callback if this is the first time the
-            # device has been seen or if an advertising data property changed.
-            # Otherwise we get a flood of callbacks from RSSI changing.
-            if first_time_seen or not _ADVERTISING_DATA_PROPERTIES.isdisjoint(changed):
-                # TODO: this should be deep copy, not shallow
-                callback(device_path, cast(Device1, device.copy()))
+_global_instances: MutableMapping[Any, BlueZManager] = WeakKeyDictionary()
 
 
 async def get_global_bluez_manager() -> BlueZManager:
     """
-    Gets the initialized global BlueZ manager instance.
+    Gets an existing initialized global BlueZ manager instance associated with the current event loop,
+    or initializes a new instance.
     """
 
-    if not hasattr(get_global_bluez_manager, "instance"):
-        setattr(get_global_bluez_manager, "instance", BlueZManager())
-
-    instance: BlueZManager = getattr(get_global_bluez_manager, "instance")
+    loop = asyncio.get_running_loop()
+    try:
+        instance = _global_instances[loop]
+    except KeyError:
+        instance = _global_instances[loop] = BlueZManager()
 
     await instance.async_init()
 
