@@ -9,9 +9,14 @@ Created by kevincar <kevincarrolldavis@gmail.com>
 import asyncio
 import itertools
 import logging
-from typing import Callable, Any, Dict, Iterable, NewType, Optional
+import sys
+from typing import Any, Dict, Iterable, NewType, Optional
+
+if sys.version_info < (3, 11):
+    from async_timeout import timeout as async_timeout
+else:
+    from asyncio import timeout as async_timeout
 
-import async_timeout
 import objc
 from Foundation import NSNumber, NSObject, NSArray, NSData, NSError, NSUUID, NSString
 from CoreBluetooth import (
@@ -22,7 +27,8 @@ from CoreBluetooth import (
     CBCharacteristicWriteWithResponse,
 )
 
-from bleak.exc import BleakError
+from ...exc import BleakError
+from ..client import NotifyCallback
 
 # logging.basicConfig(level=logging.DEBUG)
 logger = logging.getLogger(__name__)
@@ -60,7 +66,7 @@ class PeripheralDelegate(NSObject):
         self._descriptor_write_futures: Dict[int, asyncio.Future] = {}
 
         self._characteristic_notify_change_futures: Dict[int, asyncio.Future] = {}
-        self._characteristic_notify_callbacks: Dict[int, Callable[[str, Any], Any]] = {}
+        self._characteristic_notify_callbacks: Dict[int, NotifyCallback] = {}
 
         self._read_rssi_futures: Dict[NSUUID, asyncio.Future] = {}
 
@@ -92,12 +98,12 @@ class PeripheralDelegate(NSObject):
         )
 
     @objc.python_method
-    async def discover_services(self) -> NSArray:
+    async def discover_services(self, services: Optional[NSArray]) -> NSArray:
         future = self._event_loop.create_future()
 
         self._services_discovered_future = future
         try:
-            self.peripheral.discoverServices_(None)
+            self.peripheral.discoverServices_(services)
             return await future
         finally:
             del self._services_discovered_future
@@ -145,7 +151,7 @@ class PeripheralDelegate(NSObject):
         self._characteristic_read_futures[characteristic.handle()] = future
         try:
             self.peripheral.readValueForCharacteristic_(characteristic)
-            async with async_timeout.timeout(timeout):
+            async with async_timeout(timeout):
                 return await future
         finally:
             del self._characteristic_read_futures[characteristic.handle()]
@@ -204,7 +210,7 @@ class PeripheralDelegate(NSObject):
 
     @objc.python_method
     async def start_notifications(
-        self, characteristic: CBCharacteristic, callback: Callable[[str, Any], Any]
+        self, characteristic: CBCharacteristic, callback: NotifyCallback
     ) -> None:
         c_handle = characteristic.handle()
         if c_handle in self._characteristic_notify_callbacks:
@@ -366,8 +372,9 @@ class PeripheralDelegate(NSObject):
         if not future:
             if error is None:
                 notify_callback = self._characteristic_notify_callbacks.get(c_handle)
+
                 if notify_callback:
-                    notify_callback(c_handle, bytearray(value))
+                    notify_callback(bytearray(value))
             return
 
         if error is not None:
